"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Skql = void 0;
const openapi_operation_executor_1 = require("@comake/openapi-operation-executor");
const axios_1 = __importDefault(require("axios"));
const rdf_validate_shacl_1 = __importDefault(require("rdf-validate-shacl"));
const Mapper_1 = require("./mapping/Mapper");
const MemoryQueryAdapter_1 = require("./storage/memory/MemoryQueryAdapter");
const SparqlQueryAdapter_1 = require("./storage/sparql/SparqlQueryAdapter");
const Util_1 = require("./util/Util");
const Vocabularies_1 = require("./util/Vocabularies");
const DEFAULT_MAPPING_FRAME = { '@id': Vocabularies_1.SKL.mappingSubject };
class Skql {
    constructor(options) {
        switch (options.type) {
            case 'memory':
                this.adapter = new MemoryQueryAdapter_1.MemoryQueryAdapter(options);
                break;
            case 'sparql':
                this.adapter = new SparqlQueryAdapter_1.SparqlQueryAdapter(options);
                break;
            default:
                throw new Error('No schema source found in setSchema args.');
        }
        this.inputFiles = options.inputFiles;
        this.mapper = new Mapper_1.Mapper({ functions: options.functions });
        // eslint-disable-next-line func-style
        const getVerbHandler = (getTarget, property) => async (verbArgs) => this.handleVerb(property, verbArgs);
        this.verb = new Proxy({}, { get: getVerbHandler });
    }
    async executeRawQuery(query) {
        return await this.adapter.executeRawQuery(query);
    }
    async executeRawEntityQuery(query, frame) {
        return await this.adapter.executeRawEntityQuery(query, frame);
    }
    async find(options) {
        const entity = await this.adapter.find(options);
        if (entity) {
            return entity;
        }
        throw new Error(`No schema found with fields matching ${JSON.stringify(options)}`);
    }
    async findBy(where) {
        const entity = await this.adapter.findBy(where);
        if (entity) {
            return entity;
        }
        throw new Error(`No schema found with fields matching ${JSON.stringify(where)}`);
    }
    async findAll(options) {
        return await this.adapter.findAll(options);
    }
    async findAllBy(where) {
        return await this.adapter.findAllBy(where);
    }
    async exists(where) {
        return await this.adapter.exists(where);
    }
    async count(where) {
        return await this.adapter.count(where);
    }
    async save(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return await this.adapter.save(entityOrEntities);
        }
        return await this.adapter.save(entityOrEntities);
    }
    async destroy(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return await this.adapter.destroy(entityOrEntities);
        }
        return await this.adapter.destroy(entityOrEntities);
    }
    async destroyAll() {
        return await this.adapter.destroyAll();
    }
    async performMapping(args, mapping, frame) {
        const nonReferenceMappings = await this.resolveMappingReferences(mapping);
        return await this.mapper.apply(args, nonReferenceMappings, frame ?? DEFAULT_MAPPING_FRAME);
    }
    async performMappingAndConvertToJSON(args, mapping, convertToJsonDeep = true, frame) {
        const nonReferenceMappings = await this.resolveMappingReferences(mapping);
        const jsonLd = await this.mapper.applyAndFrameSklProperties(args, nonReferenceMappings, frame ?? DEFAULT_MAPPING_FRAME);
        return (0, Util_1.toJSON)(jsonLd, convertToJsonDeep);
    }
    async handleVerb(verbName, verbArgs) {
        const verb = await this.findVerbWithName(verbName);
        if (verbArgs.noun) {
            return this.handleNounMappingVerb(verb, verbArgs);
        }
        if (verbArgs.account) {
            return this.handleIntegrationVerb(verb, verbArgs);
        }
        throw new Error(`Verb parameters must include either a noun or an account.`);
    }
    async findVerbWithName(verbName) {
        const verbSchemaId = (0, Util_1.constructUri)(Vocabularies_1.sklNamespace, verbName);
        try {
            return await this.findBy({ id: verbSchemaId });
        }
        catch {
            throw new Error(`Failed to find the verb ${verbName} in the schema.`);
        }
    }
    async handleIntegrationVerb(verb, args) {
        await this.assertVerbParamsMatchParameterSchemas(args, verb);
        const account = await this.findBy({ id: args.account });
        const integrationId = account[Vocabularies_1.SKL.integration]['@id'];
        const mapping = await this.findVerbIntegrationMapping(verb['@id'], integrationId);
        const operationArgs = await this.performParameterMappingOnArgsIfDefined(args, mapping);
        const operationInfo = await this.performOperationMappingWithArgs(args, mapping);
        const rawReturnValue = await this.performOperation(operationInfo, operationArgs, account);
        if (operationInfo[Vocabularies_1.SKL.schemeName] && rawReturnValue.authorizationUrl) {
            return {
                '@type': '@json',
                '@value': rawReturnValue,
            };
        }
        if (mapping[Vocabularies_1.SKL.returnValueMapping]) {
            const mappedReturnValue = await this.performReturnValueMappingWithFrame(rawReturnValue.data, mapping, verb);
            await this.assertVerbReturnValueMatchesReturnTypeSchema(mappedReturnValue, verb);
            return mappedReturnValue;
        }
        return rawReturnValue;
    }
    async findVerbIntegrationMapping(verbId, integrationId) {
        return await this.findBy({
            type: Vocabularies_1.SKL.VerbIntegrationMapping,
            [Vocabularies_1.SKL.verb]: verbId,
            [Vocabularies_1.SKL.integration]: integrationId,
        });
    }
    async performOperationMappingWithArgs(args, mapping) {
        return await this.performMapping(args, mapping[Vocabularies_1.SKL.operationMapping]);
    }
    async performOperation(operationInfo, operationArgs, account) {
        if (operationInfo[Vocabularies_1.SKL.schemeName]) {
            return await this.performSecuritySchemeStageWithCredentials(operationInfo, operationArgs, account);
        }
        if (operationInfo[Vocabularies_1.SKL.dataSource]) {
            return await this.getDataFromDataSource((0, Util_1.getValueIfDefined)(operationInfo[Vocabularies_1.SKL.dataSource]));
        }
        if (operationInfo[Vocabularies_1.SKL.operationId]) {
            return await this.performOpenapiOperationWithCredentials((0, Util_1.getValueIfDefined)(operationInfo[Vocabularies_1.SKL.operationId]), operationArgs, account);
        }
        throw new Error('Operation not supported.');
    }
    async performReturnValueMappingWithFrame(data, mapping, verb) {
        return await this.performMapping(data, mapping[Vocabularies_1.SKL.returnValueMapping], {
            ...(0, Util_1.getValueIfDefined)(verb[Vocabularies_1.SKL.returnValueFrame]),
            ...(0, Util_1.getValueIfDefined)(mapping[Vocabularies_1.SKL.returnValueFrame]),
        });
    }
    async resolveMappingReferences(mapping) {
        if (Array.isArray(mapping)) {
            return await Promise.all(mapping.map(async (subMapping) => await this.resolveMappingReferences(subMapping)));
        }
        return mapping;
    }
    async performParameterMappingOnArgsIfDefined(args, mapping, convertToJsonDeep = true) {
        if (mapping[Vocabularies_1.SKL.parameterMapping]) {
            return await this.performMappingAndConvertToJSON(args, mapping[Vocabularies_1.SKL.parameterMapping], convertToJsonDeep);
        }
        return args;
    }
    async getOpenApiDescriptionForIntegration(integrationId) {
        const openApiDescriptionSchema = await this.findBy({
            type: Vocabularies_1.SKL.OpenApiDescription,
            [Vocabularies_1.SKL.integration]: integrationId,
        });
        return (0, Util_1.getValueIfDefined)(openApiDescriptionSchema[Vocabularies_1.SKL.openApiDescription]);
    }
    async findSecurityCredentialsForAccount(accountId) {
        return await this.findBy({
            type: Vocabularies_1.SKL.SecurityCredentials,
            [Vocabularies_1.SKL.account]: accountId,
        });
    }
    async findSecurityCredentialsForAccountIfDefined(accountId) {
        try {
            return await this.findSecurityCredentialsForAccount(accountId);
        }
        catch {
            return undefined;
        }
    }
    async createOpenApiOperationExecutorWithSpec(openApiDescription) {
        const executor = new openapi_operation_executor_1.OpenApiOperationExecutor();
        await executor.setOpenapiSpec(openApiDescription);
        return executor;
    }
    async handleNounMappingVerb(verb, args) {
        const mapping = await this.findVerbNounMapping(verb['@id'], args.noun);
        if (mapping[Vocabularies_1.SKL.mapping]) {
            return await this.performMapping(args, mapping[Vocabularies_1.SKL.mapping], {
                ...(0, Util_1.getValueIfDefined)(verb[Vocabularies_1.SKL.returnValueFrame]),
                ...(0, Util_1.getValueIfDefined)(mapping[Vocabularies_1.SKL.returnValueFrame]),
            });
        }
        const verbArgs = await this.performParameterMappingOnArgsIfDefined(args, mapping, false);
        const verbInfoJsonLd = await this.performVerbMappingWithArgs(args, mapping);
        const mappedVerb = await this.findBy({
            id: (0, Util_1.getValueIfDefined)(verbInfoJsonLd[Vocabularies_1.SKL.verb]),
        });
        return this.handleIntegrationVerb(mappedVerb, verbArgs);
    }
    async findVerbNounMapping(verbId, noun) {
        return await this.findBy({
            type: Vocabularies_1.SKL.VerbNounMapping,
            [Vocabularies_1.SKL.verb]: verbId,
            [Vocabularies_1.SKL.noun]: noun,
        });
    }
    async performVerbMappingWithArgs(args, mapping) {
        return await this.performMapping(args, mapping[Vocabularies_1.SKL.verbMapping]);
    }
    async assertVerbParamsMatchParameterSchemas(verbParams, verb) {
        const verbParamsAsJsonLd = {
            '@context': (0, Util_1.getValueIfDefined)(verb[Vocabularies_1.SKL.parametersContext]),
            '@type': Vocabularies_1.SKL.Parameters,
            ...verbParams,
        };
        const parametersSchema = verb[Vocabularies_1.SKL.parameters];
        const report = await this.convertToQuadsAndValidateAgainstShape(verbParamsAsJsonLd, parametersSchema);
        if (!report.conforms) {
            throw new Error(`${verb[Vocabularies_1.RDFS.label]} parameters do not conform to the schema`);
        }
    }
    async performOpenapiOperationWithCredentials(operationId, operationArgs, account) {
        const integrationId = account[Vocabularies_1.SKL.integration]['@id'];
        const openApiDescription = await this.getOpenApiDescriptionForIntegration(integrationId);
        const openApiExecutor = await this.createOpenApiOperationExecutorWithSpec(openApiDescription);
        const securityCredentials = await this.findSecurityCredentialsForAccountIfDefined(account['@id']);
        const configuration = {
            accessToken: securityCredentials
                ? (0, Util_1.getValueIfDefined)(securityCredentials[Vocabularies_1.SKL.accessToken])
                : undefined,
            apiKey: securityCredentials
                ? (0, Util_1.getValueIfDefined)(securityCredentials[Vocabularies_1.SKL.apiKey])
                : undefined,
            basePath: (0, Util_1.getValueIfDefined)(account[Vocabularies_1.SKL.overrideBasePath]),
        };
        return await openApiExecutor.executeOperation(operationId, configuration, operationArgs)
            .catch(async (error) => {
            if (axios_1.default.isAxiosError(error) && await this.isInvalidTokenError(error, integrationId) && securityCredentials) {
                const refreshedConfiguration = await this.refreshOpenApiToken(securityCredentials, openApiExecutor, integrationId);
                return await openApiExecutor.executeOperation(operationId, refreshedConfiguration, operationArgs);
            }
            throw error;
        });
    }
    async isInvalidTokenError(error, integrationId) {
        const integration = await this.findBy({ id: integrationId });
        const errorMatcher = (0, Util_1.getValueIfDefined)(integration[Vocabularies_1.SKL.invalidTokenErrorMatcher]);
        if (errorMatcher && (error.response?.status === errorMatcher.status)) {
            if (!errorMatcher.messageRegex) {
                return true;
            }
            if (error.response?.statusText &&
                new RegExp(errorMatcher.messageRegex, 'u').test(error.response?.statusText)) {
                return true;
            }
        }
        return false;
    }
    async refreshOpenApiToken(securityCredentialsSchema, openApiExecutor, integrationId) {
        const getOauthTokenVerb = await this.findBy({ id: Vocabularies_1.SKL.getOauthTokens });
        const mapping = await this.findVerbIntegrationMapping(getOauthTokenVerb['@id'], integrationId);
        const operationArgs = await this.performParameterMappingOnArgsIfDefined({ refreshToken: (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.refreshToken]) }, mapping);
        const operationInfoJsonLd = await this.performOperationMappingWithArgs({}, mapping);
        const configuration = this.getConfigurationFromSecurityCredentials(securityCredentialsSchema);
        const rawReturnValue = await openApiExecutor.executeSecuritySchemeStage((0, Util_1.getValueIfDefined)(operationInfoJsonLd[Vocabularies_1.SKL.schemeName]), (0, Util_1.getValueIfDefined)(operationInfoJsonLd[Vocabularies_1.SKL.oauthFlow]), (0, Util_1.getValueIfDefined)(operationInfoJsonLd[Vocabularies_1.SKL.stage]), configuration, operationArgs);
        const mappedReturnValue = await this.performReturnValueMappingWithFrame(rawReturnValue.data, mapping, getOauthTokenVerb);
        await this.assertVerbReturnValueMatchesReturnTypeSchema(mappedReturnValue, getOauthTokenVerb);
        securityCredentialsSchema[Vocabularies_1.SKL.accessToken] = (0, Util_1.getValueIfDefined)(mappedReturnValue[Vocabularies_1.SKL.accessToken]);
        securityCredentialsSchema[Vocabularies_1.SKL.refreshToken] = (0, Util_1.getValueIfDefined)(mappedReturnValue[Vocabularies_1.SKL.refreshToken]);
        await this.save(securityCredentialsSchema);
        return { accessToken: (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.accessToken]) };
    }
    getConfigurationFromSecurityCredentials(securityCredentialsSchema) {
        const username = (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.clientId]);
        const password = (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.clientSecret]);
        const accessToken = (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.accessToken]);
        return { username, password, accessToken };
    }
    async assertVerbReturnValueMatchesReturnTypeSchema(returnValue, verb) {
        const returnTypeSchemaObject = await this.getReturnTypeSchemaFromVerb(verb);
        let report;
        if (returnValue && Object.keys(returnValue).length > 0 && returnTypeSchemaObject) {
            returnTypeSchemaObject[Vocabularies_1.SHACL.targetNode] = { '@id': returnValue['@id'] };
            report = await this.convertToQuadsAndValidateAgainstShape(returnValue, returnTypeSchemaObject);
        }
        if (report && !report?.conforms) {
            throw new Error(`Return value ${returnValue['@id']} does not conform to the schema`);
        }
    }
    async getReturnTypeSchemaFromVerb(verb) {
        const returnTypeSchema = verb[Vocabularies_1.SKL.returnValue];
        if (typeof returnTypeSchema === 'object' && returnTypeSchema['@type']) {
            return returnTypeSchema;
        }
        if (typeof returnTypeSchema === 'object' && returnTypeSchema['@id']) {
            return await this.findBy({ id: returnTypeSchema['@id'] });
        }
        throw new Error('returnTypeSchema is not properly formatted.');
    }
    async convertToQuadsAndValidateAgainstShape(value, shape) {
        const valueAsQuads = await (0, Util_1.convertJsonLdToQuads)([value]);
        const shapeQuads = await (0, Util_1.convertJsonLdToQuads)(shape);
        const validator = new rdf_validate_shacl_1.default(shapeQuads);
        return validator.validate(valueAsQuads);
    }
    async performSecuritySchemeStageWithCredentials(operationInfo, operationArgs, account) {
        const integrationId = account[Vocabularies_1.SKL.integration]['@id'];
        const openApiDescription = await this.getOpenApiDescriptionForIntegration(integrationId);
        const securityCredentialsSchema = await this.findSecurityCredentialsForAccountIfDefined(account['@id']);
        let configuration;
        if (securityCredentialsSchema) {
            configuration = this.getConfigurationFromSecurityCredentials(securityCredentialsSchema);
            operationArgs.client_id = (0, Util_1.getValueIfDefined)(securityCredentialsSchema[Vocabularies_1.SKL.clientId]);
        }
        else {
            configuration = {};
        }
        const openApiExecutor = await this.createOpenApiOperationExecutorWithSpec(openApiDescription);
        return await openApiExecutor.executeSecuritySchemeStage((0, Util_1.getValueIfDefined)(operationInfo[Vocabularies_1.SKL.schemeName]), (0, Util_1.getValueIfDefined)(operationInfo[Vocabularies_1.SKL.oauthFlow]), (0, Util_1.getValueIfDefined)(operationInfo[Vocabularies_1.SKL.stage]), configuration, operationArgs);
    }
    async getDataFromDataSource(dataSourceId) {
        const dataSource = await this.findBy({ id: dataSourceId });
        if (dataSource['@type'] === Vocabularies_1.SKL.JsonDataSource) {
            const data = this.getDataFromJsonDataSource(dataSource);
            return { data };
        }
        throw new Error(`DataSource type ${dataSource['@type']} is not supported.`);
    }
    getDataFromJsonDataSource(dataSource) {
        if (dataSource[Vocabularies_1.SKL.source]) {
            const sourceValue = (0, Util_1.getValueIfDefined)(dataSource[Vocabularies_1.SKL.source]);
            return this.getJsonDataFromSource(sourceValue);
        }
        return (0, Util_1.getValueIfDefined)(dataSource[Vocabularies_1.SKL.data]);
    }
    getJsonDataFromSource(source) {
        if (this.inputFiles && source in this.inputFiles) {
            const file = this.inputFiles[source];
            return JSON.parse(file);
        }
        // eslint-disable-next-line unicorn/expiring-todo-comments
        // TODO add support for remote sources
        throw new Error(`Failed to get data from source ${source}`);
    }
}
exports.Skql = Skql;
//# sourceMappingURL=Skql.js.map