"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MemoryQueryAdapter = void 0;
const TripleUtil_1 = require("../util/TripleUtil");
const Util_1 = require("../util/Util");
const Vocabularies_1 = require("../util/Vocabularies");
const FindOperator_1 = require("./FindOperator");
/**
 * A {@link QueryAdapter} that stores data in memory.
 */
class MemoryQueryAdapter {
    constructor(schemas) {
        this.schemas = {};
        for (const schema of schemas) {
            this.schemas[schema['@id']] = schema;
        }
    }
    async find(options) {
        if (options?.where?.id && Object.keys(options.where).length === 1 && typeof options.where.id === 'string') {
            return this.schemas[options.where.id] ?? null;
        }
        if (options?.where) {
            for (const entity of Object.values(this.schemas)) {
                const matches = await this.entityMatchesQuery(entity, options.where);
                if (matches) {
                    return entity;
                }
            }
            return null;
        }
        return Object.values(this.schemas)[0] ?? null;
    }
    async findBy(where) {
        return this.find({ where });
    }
    async findAll(options) {
        let results = [];
        if (options?.where?.id && Object.keys(options.where).length === 1 && typeof options.where.id === 'string') {
            const schema = this.schemas[options.where.id];
            if (schema) {
                results = [schema];
            }
        }
        else if (options?.where) {
            for (const entity of Object.values(this.schemas)) {
                const matches = await this.entityMatchesQuery(entity, options.where);
                if (matches) {
                    results.push(entity);
                }
            }
        }
        else {
            results = Object.values(this.schemas);
        }
        if (options?.limit ?? options?.offset) {
            const start = options?.offset ?? 0;
            const end = options?.limit && options?.offset
                ? options.offset + options.limit
                : options?.limit ?? undefined;
            return results.slice(start, end);
        }
        return results;
    }
    async findAllBy(where) {
        return this.findAll({ where });
    }
    async entityMatchesQuery(entity, where) {
        for (const [fieldName, fieldValue] of Object.entries(where)) {
            const matches = await this.entityMatchesField(entity, fieldName, fieldValue);
            if (!matches) {
                return false;
            }
        }
        return true;
    }
    async entityMatchesField(entity, fieldName, fieldValue) {
        if (FindOperator_1.FindOperator.isFindOperator(fieldValue)) {
            return await this.handleOperator(fieldValue.operator, {
                in: async () => {
                    const values = fieldValue.value;
                    for (const valueItem of values) {
                        if (await this.entityMatchesField(entity, fieldName, valueItem)) {
                            return true;
                        }
                    }
                    return false;
                },
                not: async () => {
                    if (FindOperator_1.FindOperator.isFindOperator(fieldValue.value)) {
                        return !await this.entityMatchesField(entity, fieldName, fieldValue.value);
                    }
                    const valueItem = fieldValue.value;
                    return !await this.entityMatchesField(entity, fieldName, valueItem);
                },
                equal: async () => {
                    const valueItem = fieldValue.value;
                    return this.entityMatchesField(entity, fieldName, valueItem);
                },
            });
        }
        if (fieldName === 'id') {
            return entity['@id'] === fieldValue;
        }
        if (fieldName === 'type') {
            return this.isInstanceOf(entity, fieldValue);
        }
        if (Array.isArray(fieldValue)) {
            for (const valueItem of fieldValue) {
                if (!await this.entityMatchesField(entity, fieldName, valueItem)) {
                    return false;
                }
            }
            return true;
        }
        if (typeof fieldValue === 'object') {
            if (Array.isArray(entity[fieldName])) {
                for (const subFieldValue of entity[fieldName]) {
                    const matches = await this.findOptionWhereMatchesNodeObject(fieldValue, subFieldValue);
                    if (matches) {
                        return true;
                    }
                }
                return false;
            }
            if (typeof entity[fieldName] === 'object') {
                return await this.findOptionWhereMatchesNodeObject(fieldValue, entity[fieldName]);
            }
            return false;
        }
        if (Array.isArray(entity[fieldName])) {
            return entity[fieldName].some((field) => this.fieldValueMatchesField(fieldValue, field));
        }
        return this.fieldValueMatchesField(fieldValue, entity[fieldName]);
    }
    fieldValueMatchesField(fieldValue, field) {
        if (typeof field === 'object') {
            if (field['@id']) {
                return field['@id'] === fieldValue;
            }
            if (field['@value']) {
                const jsValue = (0, TripleUtil_1.toJSValueFromDataType)(field['@value'], field['@type']);
                return jsValue === fieldValue;
            }
        }
        return field === fieldValue;
    }
    async findOptionWhereMatchesNodeObject(fieldValue, nodeObject) {
        if (nodeObject['@id'] && Object.keys(nodeObject).length === 1) {
            const subEntity = await this.findBy({ id: nodeObject['@id'] });
            if (subEntity) {
                return this.entityMatchesQuery(subEntity, fieldValue);
            }
            return false;
        }
        return this.entityMatchesQuery(nodeObject, fieldValue);
    }
    async handleOperator(operator, operatorHandlers) {
        if (operator in operatorHandlers) {
            return await operatorHandlers[operator]();
        }
        throw new Error(`Unsupported operator "${operator}"`);
    }
    isInstanceOf(entity, targetClass) {
        const classes = this.getSubClassesOf(targetClass);
        const entityTypes = (0, Util_1.ensureArray)(entity['@type']);
        return entityTypes.some((type) => classes.includes(type));
    }
    getSubClassesOf(targetClass) {
        // Cache subclassesOf
        return Object.values(this.schemas).reduce((subClasses, schema) => {
            const subClassOf = (0, Util_1.ensureArray)(schema[Vocabularies_1.RDFS.subClassOf]);
            const isSubClassOfTarget = subClassOf.some((subClass) => subClass['@id'] === targetClass);
            if (isSubClassOfTarget) {
                subClasses = [
                    ...subClasses,
                    schema['@id'],
                    ...this.getSubClassesOf(schema['@id']),
                ];
            }
            return subClasses;
        }, [targetClass]);
    }
    async exists(where) {
        const res = await this.find({ where });
        return res !== null;
    }
    async save(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return entityOrEntities.map((entity) => this.saveEntity(entity));
        }
        return this.saveEntity(entityOrEntities);
    }
    saveEntity(entity) {
        const savedEntity = { ...entity };
        this.schemas[entity['@id']] = savedEntity;
        return savedEntity;
    }
    async destroy(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return entityOrEntities.map((entity) => this.destroyEntity(entity));
        }
        return this.destroyEntity(entityOrEntities);
    }
    destroyEntity(entity) {
        const existingEntity = this.schemas[entity['@id']];
        if (!existingEntity) {
            throw new Error(`Entity with id ${entity['@id']} does not exist.`);
        }
        // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
        delete this.schemas[existingEntity['@id']];
        return existingEntity;
    }
}
exports.MemoryQueryAdapter = MemoryQueryAdapter;
//# sourceMappingURL=MemoryQueryAdapter.js.map