"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SparqlQueryAdapter = void 0;
/* eslint-disable @typescript-eslint/naming-convention */
const data_model_1 = __importDefault(require("@rdfjs/data-model"));
const sparql_http_client_1 = __importDefault(require("sparql-http-client"));
const sparqljs_1 = require("sparqljs");
const Util_1 = require("../util/Util");
const VariableGenerator_1 = require("../util/VariableGenerator");
const Vocabularies_1 = require("../util/Vocabularies");
/**
 * A {@link QueryAdapter} that stores data in a database through a sparql endpoint.
 */
class SparqlQueryAdapter {
    constructor(endpointUrl) {
        this.sparqlClient = new sparql_http_client_1.default({ endpointUrl });
        this.sparqlGenerator = new sparqljs_1.Generator();
    }
    async find(options) {
        const variableGenerator = new VariableGenerator_1.VariableGenerator();
        const graph = data_model_1.default.variable('graph');
        const subject = data_model_1.default.variable('subject');
        const predicate = data_model_1.default.variable('predicate');
        const object = data_model_1.default.variable('object');
        const namedGraphQueryData = this.createWhereQueryData(graph, variableGenerator, options?.where);
        const graphQuery = {
            type: 'query',
            queryType: 'SELECT',
            variables: [graph],
            where: this.buildPatternsFromQueryData(graph, namedGraphQueryData),
            limit: 1,
        };
        const queryJson = {
            type: 'query',
            queryType: 'SELECT',
            prefixes: {},
            variables: [subject, predicate, object],
            where: [
                {
                    type: 'graph',
                    name: graph,
                    patterns: [
                        {
                            type: 'bgp',
                            triples: [
                                { subject, predicate, object },
                            ],
                        },
                    ],
                },
                graphQuery,
            ],
        };
        const generatedQuery = this.sparqlGenerator.stringify(queryJson);
        const response = await this.executeSparqlQueryAndGetData(generatedQuery);
        return response;
    }
    buildPatternsFromQueryData(graph, queryData) {
        const basicGraphPatterns = queryData.triples.length === 0 ? [] : [{ type: 'bgp', triples: queryData.triples }];
        const patterns = [
            {
                type: 'graph',
                // File a bug in sparqljs types. GraphPattern should accept Variables in name
                name: graph,
                patterns: basicGraphPatterns,
            },
        ];
        if (queryData.filters.length === 1) {
            patterns.push({
                type: 'filter',
                expression: queryData.filters[0],
            });
        }
        else if (queryData.filters.length > 1) {
            patterns.push({
                type: 'filter',
                expression: {
                    type: 'operation',
                    operator: '&&',
                    args: queryData.filters,
                },
            });
        }
        return patterns;
    }
    createWhereQueryData(parentVariable, variableGenerator, where) {
        if (!where) {
            return {
                triples: [{
                        subject: data_model_1.default.variable(variableGenerator.getNext()),
                        predicate: data_model_1.default.variable(variableGenerator.getNext()),
                        object: data_model_1.default.variable(variableGenerator.getNext()),
                    }],
                filters: [],
            };
        }
        let triples = [];
        let filters = [];
        for (const [key, value] of Object.entries(where)) {
            const graphQueryData = this.createWhereQueryDataFromKeyValue(key, value, parentVariable, variableGenerator);
            triples = [...triples, ...graphQueryData.triples];
            filters = [...filters, ...graphQueryData.filters];
        }
        return { triples, filters };
    }
    async findBy(where) {
        return this.find({ where });
    }
    async findAll(options) {
        return [];
    }
    async findAllBy(where) {
        return this.findAll({ where });
    }
    async save(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return [];
        }
        return { '@id': '', '@type': '' };
    }
    async destroy(entityOrEntities) {
        if (Array.isArray(entityOrEntities)) {
            return [];
        }
        return { '@id': '', '@type': '' };
    }
    // Remove
    // private generateSparqlQueryStringFromFindOneOptions(options?: FindOneOptions): string {
    //   return '';
    // }
    // private generateSparqlQueryStringFromFindAllOptions(options?: FindAllOptions): string {
    //   return '';
    // }
    async executeSparqlQueryAndGetData(query) {
        console.log('executing query', query);
        const stream = await this.sparqlClient.query.select(query);
        return new Promise((resolve, reject) => {
            const data = [];
            stream.on('data', (row) => {
                Object.entries(row).forEach((entry) => {
                    data.push(entry);
                });
            });
            stream.on('end', () => {
                resolve(data);
            });
            stream.on('error', (error) => {
                reject(error);
            });
        });
    }
    valueToLiteral(value) {
        if (typeof value === 'number') {
            return data_model_1.default.literal(value.toString(), Vocabularies_1.XSD.integer);
        }
        if (typeof value === 'boolean') {
            return data_model_1.default.literal(value.toString(), Vocabularies_1.XSD.boolean);
        }
        return data_model_1.default.literal(value.toString());
    }
    buildEqualOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '=',
            args: [leftSide, rightSide],
        };
    }
    buildTypesAndSuperTypesTriple(subject, object) {
        return {
            subject,
            predicate: Util_1.allTypesAndSuperTypesPath,
            object,
        };
    }
    createWhereQueryDataFromKeyValue(key, value, parentVariable, variableGenerator) {
        if (key === 'id') {
            return this.createWhereQueryDataForIdValue(parentVariable, value);
        }
        if (key === 'type') {
            return this.createWhereQueryDataForType(parentVariable, value);
        }
        if (typeof value === 'object') {
            return this.createWhereQueryDataForNestedWhere(key, value, parentVariable, variableGenerator);
        }
        return this.createWhereQueryDataForLiteral(key, value, parentVariable);
    }
    createWhereQueryDataForIdValue(term, value) {
        return {
            filters: [this.buildEqualOperation(term, data_model_1.default.namedNode(value))],
            triples: [],
        };
    }
    createWhereQueryDataForType(subject, value) {
        return {
            filters: [],
            triples: [this.buildTypesAndSuperTypesTriple(subject, data_model_1.default.namedNode(value))],
        };
    }
    createWhereQueryDataForNestedWhere(key, where, parentVariable, variableGenerator) {
        const subNodeVariable = data_model_1.default.variable(variableGenerator.getNext());
        const subWhereQueryData = this.createWhereQueryData(subNodeVariable, variableGenerator, where);
        return {
            filters: subWhereQueryData.filters,
            triples: [
                {
                    subject: parentVariable,
                    predicate: data_model_1.default.namedNode(key),
                    object: subNodeVariable,
                },
                ...subWhereQueryData.triples,
            ],
        };
    }
    createWhereQueryDataForLiteral(key, value, parentVariable) {
        return {
            filters: [],
            triples: [{
                    subject: parentVariable,
                    predicate: data_model_1.default.namedNode(key),
                    object: this.valueToLiteral(value),
                }],
        };
    }
}
exports.SparqlQueryAdapter = SparqlQueryAdapter;
//# sourceMappingURL=SparqlQueryAdapter.js.map