"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SparqlQueryAdapter = void 0;
const sparql_http_client_1 = __importDefault(require("sparql-http-client"));
const sparqljs_1 = require("sparqljs");
const TripleUtil_1 = require("../../util/TripleUtil");
const SparqlQueryBuilder_1 = require("./SparqlQueryBuilder");
const SparqlUpdateBuilder_1 = require("./SparqlUpdateBuilder");
/**
 * A {@link QueryAdapter} that stores data in a database through a sparql endpoint.
 */
class SparqlQueryAdapter {
    constructor(options) {
        this.sparqlClient = new sparql_http_client_1.default({
            endpointUrl: options.endpointUrl,
            updateUrl: options.updateUrl ?? options.endpointUrl,
        });
        this.sparqlGenerator = new sparqljs_1.Generator();
        this.setTimestamps = options.setTimestamps ?? false;
    }
    async executeRawQuery(query) {
        const response = await this.executeSparqlSelectAndGetData(query);
        if (response.length === 0) {
            return [];
        }
        return response
            .map((result) => Object.entries(result).reduce((obj, [key, value]) => ({
            ...obj,
            [key]: value.termType === 'Literal'
                ? (0, TripleUtil_1.toJSValueFromDataType)(value.value, value.datatype?.value)
                : value.value,
        }), {}));
    }
    async executeRawEntityQuery(query, frame) {
        const response = await this.executeSparqlSelectAndGetData(query);
        if (response.length === 0) {
            return { '@graph': [] };
        }
        return await (0, TripleUtil_1.triplesToJsonldWithFrame)(response, frame);
    }
    async find(options) {
        const queryBuilder = new SparqlQueryBuilder_1.SparqlQueryBuilder();
        const selectQueryData = queryBuilder.buildPatternsFromQueryOptions(TripleUtil_1.entityVariable, options?.where, options?.order, options?.relations);
        const entitySelectQuery = this.sparqlSelect([TripleUtil_1.entityVariable, ...selectQueryData.variables], selectQueryData.where, selectQueryData.orders, 1);
        const entitySelectGroupQuery = this.sparqlSelectGroup([entitySelectQuery]);
        selectQueryData.graphWhere.unshift(entitySelectGroupQuery);
        const query = queryBuilder.buildConstructFromEntitySelectQuery(selectQueryData.graphWhere, selectQueryData.variables, options?.select);
        const generatedQuery = this.sparqlGenerator.stringify(query);
        const responseTriples = await this.executeSparqlSelectAndGetData(generatedQuery);
        if (responseTriples.length === 0) {
            return null;
        }
        const jsonld = await (0, TripleUtil_1.triplesToJsonld)(responseTriples, options?.relations);
        return jsonld;
    }
    sparqlSelect(variables, where, order, limit, offset) {
        return {
            type: 'query',
            queryType: 'SELECT',
            variables,
            where,
            order: order.length > 0 ? order : undefined,
            limit,
            offset,
            prefixes: {},
        };
    }
    async findBy(where) {
        return this.find({ where });
    }
    async findAll(options) {
        const queryBuilder = new SparqlQueryBuilder_1.SparqlQueryBuilder();
        const selectQueryData = queryBuilder.buildPatternsFromQueryOptions(TripleUtil_1.entityVariable, options?.where, options?.order, options?.relations);
        const entitySelectVariables = [TripleUtil_1.entityVariable, ...selectQueryData.variables];
        const entitySelectQuery = this.sparqlSelect(entitySelectVariables, selectQueryData.where, selectQueryData.orders, options?.limit, options?.offset);
        let orderedEntityVariableValues;
        if (selectQueryData.orders.length > 0 && options?.limit !== 1) {
            // We need to execute the entity select query here first to get ordered results.
            const generatedEntitySelectQuery = this.sparqlGenerator.stringify(entitySelectQuery);
            const entitySelectResponse = await this.executeSparqlSelectAndGetData(generatedEntitySelectQuery);
            const valuesByVariable = this.groupByKey(entitySelectResponse);
            orderedEntityVariableValues = valuesByVariable[TripleUtil_1.entityVariable.value]
                .map((namedNode) => namedNode.value);
            const variableValueFilter = queryBuilder.buildInFilterForVariables(valuesByVariable);
            selectQueryData.graphWhere.push(variableValueFilter);
        }
        else {
            const entitySelectGroupQuery = this.sparqlSelectGroup([entitySelectQuery]);
            selectQueryData.graphWhere.unshift(entitySelectGroupQuery);
        }
        const query = queryBuilder.buildConstructFromEntitySelectQuery(selectQueryData.graphWhere, selectQueryData.variables, options?.select);
        const generatedQuery = this.sparqlGenerator.stringify(query);
        const responseTriples = await this.executeSparqlSelectAndGetData(generatedQuery);
        if (responseTriples.length === 0) {
            return [];
        }
        const jsonld = await (0, TripleUtil_1.triplesToJsonld)(responseTriples, options?.relations, orderedEntityVariableValues);
        if (Array.isArray(jsonld)) {
            return jsonld;
        }
        return [jsonld];
    }
    groupByKey(entitySelectResponse) {
        return entitySelectResponse
            .reduce((obj, result) => {
            for (const [key, value] of Object.entries(result)) {
                if (!(key in obj)) {
                    obj[key] = [value];
                }
                else {
                    obj[key].push(value);
                }
            }
            return obj;
        }, {});
    }
    sparqlSelectGroup(patterns) {
        return {
            type: 'group',
            patterns,
        };
    }
    async findAllBy(where) {
        return this.findAll({ where });
    }
    async exists(where) {
        const queryBuilder = new SparqlQueryBuilder_1.SparqlQueryBuilder();
        const selectQueryData = queryBuilder.buildPatternsFromQueryOptions(TripleUtil_1.entityVariable, where);
        const query = this.sparqlAsk(selectQueryData.where);
        return await this.executeAskQueryAndGetResponse(query);
    }
    sparqlAsk(where) {
        return {
            type: 'query',
            queryType: 'ASK',
            where,
            prefixes: {},
        };
    }
    async count(where) {
        const queryBuilder = new SparqlQueryBuilder_1.SparqlQueryBuilder();
        const selectQueryData = queryBuilder.buildPatternsFromQueryOptions(TripleUtil_1.entityVariable, where);
        const query = this.sparqlCountSelect(selectQueryData.where, selectQueryData.graphWhere);
        return await this.executeSelectCountAndGetResponse(query);
    }
    sparqlCountSelect(where, graphWhere) {
        return {
            type: 'query',
            queryType: 'SELECT',
            variables: [{
                    expression: {
                        type: 'aggregate',
                        aggregation: 'count',
                        distinct: true,
                        expression: TripleUtil_1.entityVariable,
                    },
                    variable: TripleUtil_1.countVariable,
                }],
            where: [
                this.sparqlSelectGraph(TripleUtil_1.entityVariable, where),
                ...graphWhere,
            ],
            prefixes: {},
        };
    }
    sparqlSelectGraph(name, patterns) {
        return {
            type: 'graph',
            name: name,
            patterns,
        };
    }
    async save(entityOrEntities) {
        const queryBuilder = new SparqlUpdateBuilder_1.SparqlUpdateBuilder({ setTimestamps: this.setTimestamps });
        const query = queryBuilder.buildUpdate(entityOrEntities);
        await this.executeSparqlUpdate(query);
        return entityOrEntities;
    }
    async destroy(entityOrEntities) {
        const queryBuilder = new SparqlUpdateBuilder_1.SparqlUpdateBuilder();
        const query = queryBuilder.buildDelete(entityOrEntities);
        await this.executeSparqlUpdate(query);
        return entityOrEntities;
    }
    async destroyAll() {
        const queryBuilder = new SparqlUpdateBuilder_1.SparqlUpdateBuilder();
        const query = queryBuilder.buildDeleteAll();
        await this.executeSparqlUpdate(query);
    }
    async executeSparqlSelectAndGetData(query) {
        const stream = await this.sparqlClient.query.select(query);
        return new Promise((resolve, reject) => {
            const data = [];
            stream.on('data', (row) => {
                data.push(row);
            });
            stream.on('end', () => {
                resolve(data);
            });
            stream.on('error', (error) => {
                reject(error);
            });
        });
    }
    async executeSparqlUpdate(query) {
        const generatedQuery = this.sparqlGenerator.stringify(query);
        await this.sparqlClient.query.update(generatedQuery);
    }
    async executeAskQueryAndGetResponse(query) {
        const generatedQuery = this.sparqlGenerator.stringify(query);
        return await this.sparqlClient.query.ask(generatedQuery);
    }
    async executeSelectCountAndGetResponse(query) {
        const generatedQuery = this.sparqlGenerator.stringify(query);
        const stream = await this.sparqlClient.query.select(generatedQuery);
        return new Promise((resolve, reject) => {
            let countValue;
            stream.on('data', (row) => {
                const { count } = row;
                countValue = Number.parseInt(count.value, 10);
            });
            stream.on('end', () => {
                resolve(countValue);
            });
            stream.on('error', (error) => {
                reject(error);
            });
        });
    }
}
exports.SparqlQueryAdapter = SparqlQueryAdapter;
//# sourceMappingURL=SparqlQueryAdapter.js.map