"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SparqlQueryBuilder = void 0;
/* eslint-disable unicorn/expiring-todo-comments */
const data_model_1 = __importDefault(require("@rdfjs/data-model"));
const TripleUtil_1 = require("../../util/TripleUtil");
const Util_1 = require("../../util/Util");
const Vocabularies_1 = require("../../util/Vocabularies");
const FindOperator_1 = require("../FindOperator");
const VariableGenerator_1 = require("./VariableGenerator");
class SparqlQueryBuilder {
    constructor() {
        this.variableGenerator = new VariableGenerator_1.VariableGenerator();
    }
    buildInFilterForVariables(valuesByVariable) {
        return this.filterPatternFromFilters(Object.entries(valuesByVariable).map(([variableName, values]) => this.buildInOperation(data_model_1.default.variable(variableName), values)));
    }
    filterWithExpression(expression) {
        return { type: 'filter', expression };
    }
    buildPatternsFromQueryOptions(subject, where, order, relations) {
        const queryData = this.createQueryData(subject, where, order, relations);
        const wherePatterns = this.createWherePatternsFromQueryData(queryData.triples, queryData.optionalTriples, queryData.filters);
        const graphWherePatterns = this.createWherePatternsFromQueryData(queryData.graphTriples, queryData.optionalGraphTriples, queryData.graphFilters);
        return {
            variables: queryData.relationsVariables,
            where: wherePatterns,
            orders: queryData.orders,
            graphWhere: graphWherePatterns,
        };
    }
    createWherePatternsFromQueryData(triples, optionalTriples, filters) {
        const patterns = [];
        if (triples.length > 0) {
            patterns.push(this.sparqlBasicGraphPattern(triples));
        }
        if (optionalTriples.length > 0) {
            patterns.push(this.sparqlOptionalWithTriples(optionalTriples));
        }
        if (filters.length > 0) {
            patterns.push(this.filterPatternFromFilters(filters));
        }
        return patterns;
    }
    filterPatternFromFilters(filters) {
        if (filters.length > 1) {
            return this.filterWithExpression({
                type: 'operation',
                operator: '&&',
                args: filters,
            });
        }
        return this.filterWithExpression(filters[0]);
    }
    sparqlOptionalWithTriples(triples) {
        return {
            type: 'optional',
            patterns: [this.sparqlBasicGraphPattern(triples)],
        };
    }
    sparqlBasicGraphPattern(triples) {
        return { type: 'bgp', triples };
    }
    buildConstructFromEntitySelectQuery(graphWhere, graphSelectVariables, select) {
        let triples;
        let where = [];
        if (select) {
            // TODO: fix when select and relations are used.
            triples = this.createSelectPattern(select, TripleUtil_1.entityVariable);
            where = [
                this.sparqlOptionalSelectGraph(TripleUtil_1.entityVariable, triples),
                ...graphWhere
            ];
        }
        else {
            const graphSelectsAndTriplePatterns = this.createGraphSelectsAndTriplePatterns(graphSelectVariables);
            const entityGraphTriple = { subject: TripleUtil_1.subjectNode, predicate: TripleUtil_1.predicateNode, object: TripleUtil_1.objectNode };
            triples = [
                entityGraphTriple,
                ...graphSelectsAndTriplePatterns.triples,
            ];
            where = [
                ...graphWhere,
                this.sparqlSelectGraph(TripleUtil_1.entityVariable, [this.sparqlBasicGraphPattern([entityGraphTriple])]),
                ...graphSelectsAndTriplePatterns.graphPatterns,
            ];
        }
        return {
            type: 'query',
            prefixes: {},
            queryType: 'CONSTRUCT',
            template: triples,
            where,
        };
    }
    createSelectPattern(select, subject) {
        if (Array.isArray(select)) {
            return select.map((selectPredicate) => ({
                subject,
                predicate: data_model_1.default.namedNode(selectPredicate),
                object: this.createVariable(),
            }));
        }
        return Object.entries(select).reduce((arr, [key, value]) => {
            const variable = this.createVariable();
            arr.push({ subject, predicate: data_model_1.default.namedNode(key), object: variable });
            if (typeof value === 'object') {
                arr = [...arr, ...this.createSelectPattern(value, variable)];
            }
            return arr;
        }, []);
    }
    createGraphSelectsAndTriplePatterns(variables) {
        return variables.reduce((obj, variable) => {
            const triple = {
                subject: this.createVariable(),
                predicate: this.createVariable(),
                object: this.createVariable(),
            };
            obj.triples.push(triple);
            obj.graphPatterns.push({
                type: 'optional',
                patterns: [
                    this.sparqlSelectGraph(variable, [this.sparqlBasicGraphPattern([triple])]),
                ],
            });
            return obj;
        }, { triples: [], graphPatterns: [] });
    }
    sparqlSelectGraph(name, patterns) {
        return {
            type: 'graph',
            name: name,
            patterns,
        };
    }
    sparqlOptionalSelectGraph(name, triples) {
        return {
            type: 'graph',
            name: name,
            patterns: [{
                    type: 'optional',
                    patterns: [{
                            type: 'bgp',
                            triples,
                        }],
                }],
        };
    }
    createQueryData(subject, where, order, relations) {
        const whereQueryData = this.createWhereQueryData(subject, where);
        const orderQueryData = this.createOrderQueryData(subject, order);
        const relationsQueryData = this.createRelationsQueryData(subject, relations);
        return {
            relationsVariables: relationsQueryData.variables,
            triples: whereQueryData.triples,
            filters: whereQueryData.filters,
            optionalTriples: [...relationsQueryData.triples, ...orderQueryData.triples],
            orders: orderQueryData.orders,
            graphTriples: whereQueryData.graphTriples,
            optionalGraphTriples: [...relationsQueryData.graphTriples],
            graphFilters: [...whereQueryData.graphFilters],
        };
    }
    createWhereQueryData(subject, where) {
        const hasSingleKey = Object.keys(where ?? {}).length === 1;
        const whereQueryData = Object.entries(where ?? {}).reduce((obj, [key, value]) => {
            const whereQueryDataForField = this.createWhereQueryDataForField(subject, key, value, hasSingleKey);
            return {
                triples: [...obj.triples, ...whereQueryDataForField.triples],
                graphTriples: [...obj.graphTriples, ...whereQueryDataForField.graphTriples],
                filters: [...obj.filters, ...whereQueryDataForField.filters],
                graphFilters: [...obj.graphFilters, ...whereQueryDataForField.graphFilters],
            };
        }, { triples: [], graphTriples: [], filters: [], graphFilters: [] });
        if (whereQueryData.triples.length === 0) {
            whereQueryData.triples.push({
                subject,
                predicate: this.createVariable(),
                object: this.createVariable(),
            });
        }
        return whereQueryData;
    }
    createWhereQueryDataForField(subject, field, value, isOnlyField) {
        if (field === 'id') {
            return this.createWhereQueryDataForIdValue(subject, value, isOnlyField);
        }
        if (field === 'type') {
            return this.createWhereQueryDataForType(subject, value);
        }
        const predicate = data_model_1.default.namedNode(field);
        return this.createWhereQueryDataFromKeyValue(subject, predicate, value);
    }
    createWhereQueryDataForIdValue(term, value, qualifyWithTriple) {
        let filter;
        if (FindOperator_1.FindOperator.isFindOperator(value)) {
            filter = this.resolveFindOperatorAsExpressionForId(term, value);
        }
        else {
            filter = this.buildEqualOperation(term, data_model_1.default.namedNode(value));
        }
        const queryData = {
            filters: [filter],
            graphFilters: [],
            triples: [],
            graphTriples: [],
            optionalTriples: [],
            orders: [],
        };
        if (qualifyWithTriple) {
            queryData.triples.push({
                subject: term,
                predicate: this.createVariable(),
                object: this.createVariable(),
            });
        }
        return queryData;
    }
    createWhereQueryDataForType(subject, value) {
        if (FindOperator_1.FindOperator.isFindOperator(value)) {
            const variable = this.createVariable();
            const triple = this.buildTypesAndSuperTypesTriple(subject, variable);
            const operatorFilter = this.resolveFindOperatorAsExpressionWithMultipleValues(variable, value, triple);
            return {
                filters: [operatorFilter],
                graphFilters: [],
                triples: [triple],
                graphTriples: [],
            };
        }
        return {
            filters: [],
            graphFilters: [],
            triples: [this.buildTypesAndSuperTypesTriple(subject, data_model_1.default.namedNode(value))],
            graphTriples: [],
        };
    }
    createWhereQueryDataFromKeyValue(subject, predicate, value) {
        if (FindOperator_1.FindOperator.isFindOperator(value)) {
            return this.createWhereQueryDataForFindOperator(subject, predicate, value);
        }
        if (Array.isArray(value)) {
            return value.reduce((obj, valueItem) => {
                const valueWhereQueryData = this.createWhereQueryDataFromKeyValue(subject, predicate, valueItem);
                return {
                    filters: [...obj.filters, ...valueWhereQueryData.filters],
                    triples: [...obj.triples, ...valueWhereQueryData.triples],
                    graphFilters: [...obj.graphFilters, ...valueWhereQueryData.graphFilters],
                    graphTriples: [...obj.graphTriples, ...valueWhereQueryData.graphTriples],
                };
            }, { filters: [], graphTriples: [], triples: [], graphFilters: [] });
        }
        if (typeof value === 'object') {
            if ('@value' in value) {
                return this.createWhereQueryDataForValueObject(subject, predicate, value);
            }
            return this.createWhereQueryDataForNestedWhere(subject, predicate, value);
        }
        const term = this.resolveValueToTerm(value);
        return {
            filters: [],
            graphFilters: [],
            triples: [{ subject, predicate, object: term }],
            graphTriples: [],
        };
    }
    inversePropertyPredicate(predicate) {
        return {
            type: 'path',
            pathType: '^',
            items: [predicate],
        };
    }
    createWhereQueryDataForFindOperator(subject, predicate, operator) {
        if (operator.operator === 'inverse') {
            const inversePredicate = this.inversePropertyPredicate(predicate);
            const inverseWhereQueryData = this.createWhereQueryDataFromKeyValue(subject, inversePredicate, operator.value);
            return {
                filters: [],
                triples: [],
                graphTriples: inverseWhereQueryData.triples,
                graphFilters: inverseWhereQueryData.filters,
            };
        }
        const variable = this.createVariable();
        const triple = { subject, predicate, object: variable };
        return {
            filters: [this.resolveFindOperatorAsExpressionWithMultipleValues(variable, operator, triple)],
            triples: [triple],
            graphTriples: [],
            graphFilters: [],
        };
    }
    createWhereQueryDataForNestedWhere(subject, predicate, where) {
        const subNodeVariable = this.createVariable();
        const subWhereQueryData = this.createWhereQueryData(subNodeVariable, where);
        return {
            filters: subWhereQueryData.filters,
            triples: [
                { subject, predicate, object: subNodeVariable },
                ...subWhereQueryData.triples,
            ],
            graphFilters: [],
            graphTriples: [],
        };
    }
    createWhereQueryDataForValueObject(subject, predicate, valueObject) {
        const term = this.valueObjectToTerm(valueObject);
        return {
            filters: [],
            triples: [{ subject, predicate, object: term }],
            graphFilters: [],
            graphTriples: [],
        };
    }
    valueObjectToTerm(valueObject) {
        let typeOrLanguage;
        let value;
        if ('@type' in valueObject && valueObject['@type'] === '@json') {
            typeOrLanguage = Vocabularies_1.RDF.JSON;
            value = JSON.stringify(valueObject['@value']);
        }
        else {
            typeOrLanguage = ('@type' in valueObject ? valueObject['@type'] : valueObject['@language']);
            value = valueObject['@value'].toString();
        }
        return data_model_1.default.literal(value, typeOrLanguage);
    }
    resolveFindOperatorAsExpressionWithMultipleValues(leftSide, operator, triple) {
        switch (operator.operator) {
            case 'in':
                return this.buildInOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'not':
                return this.buildNotOperationForMultiValued(leftSide, this.resolveValueToExpression(operator.value), triple);
            case 'equal':
                return this.buildEqualOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'gt':
                return this.buildGtOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'gte':
                return this.buildGteOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'lt':
                return this.buildLtOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'lte':
                return this.buildLteOperation(leftSide, this.resolveValueToExpression(operator.value));
            default:
                throw new Error(`Unsupported operator "${operator.operator}"`);
        }
    }
    resolveFindOperatorAsExpressionForId(leftSide, operator) {
        switch (operator.operator) {
            case 'in':
                return this.buildInOperation(leftSide, this.resolveValueToExpression(operator.value));
            case 'not':
                return this.buildNotOperationForId(leftSide, this.resolveValueToExpression(operator.value));
            case 'equal':
                return this.buildEqualOperation(leftSide, this.resolveValueToExpression(operator.value));
            default:
                throw new Error(`Unsupported operator "${operator.operator}"`);
        }
    }
    resolveValueToExpression(value) {
        if (FindOperator_1.FindOperator.isFindOperator(value)) {
            return value;
        }
        if (Array.isArray(value)) {
            return value.map((valueItem) => this.resolveValueToTerm(valueItem));
        }
        return this.resolveValueToTerm(value);
    }
    buildInOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: 'in',
            args: [leftSide, rightSide],
        };
    }
    buildNotInOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: 'notin',
            args: [leftSide, rightSide],
        };
    }
    buildNotOperationForMultiValued(leftSide, rightSide, triple) {
        let filterExpression;
        const rightSideIsOperation = typeof rightSide === 'object' && 'operator' in rightSide;
        if (rightSideIsOperation) {
            let expression;
            try {
                expression = this.resolveFindOperatorAsExpressionWithMultipleValues(leftSide, rightSide, triple);
            }
            catch {
                throw new Error(`Unsupported Not sub operator "${rightSide.operator}"`);
            }
            filterExpression = this.filterWithExpression(expression);
        }
        else {
            filterExpression = this.filterWithExpression(this.buildEqualOperation(leftSide, rightSide));
        }
        return {
            type: 'operation',
            operator: 'notexists',
            args: [
                {
                    type: 'group',
                    patterns: [
                        {
                            type: 'bgp',
                            triples: [triple],
                        },
                        filterExpression,
                    ],
                },
            ],
        };
    }
    buildNotOperationForId(leftSide, rightSide) {
        if (FindOperator_1.FindOperator.isFindOperator(rightSide)) {
            switch (rightSide.operator) {
                case 'in':
                    return this.buildNotInOperation(leftSide, this.resolveValueToExpression(rightSide.value));
                case 'equal':
                    return this.buildNotEqualOperation(leftSide, this.resolveValueToExpression(rightSide.value));
                default:
                    throw new Error(`Unsupported Not sub operator "${rightSide.operator}"`);
            }
        }
        return this.buildNotEqualOperation(leftSide, rightSide);
    }
    buildEqualOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '=',
            args: [leftSide, rightSide],
        };
    }
    buildGtOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '>',
            args: [leftSide, rightSide],
        };
    }
    buildGteOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '>=',
            args: [leftSide, rightSide],
        };
    }
    buildLtOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '<',
            args: [leftSide, rightSide],
        };
    }
    buildLteOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '<=',
            args: [leftSide, rightSide],
        };
    }
    buildNotEqualOperation(leftSide, rightSide) {
        return {
            type: 'operation',
            operator: '!=',
            args: [leftSide, rightSide],
        };
    }
    buildTypesAndSuperTypesTriple(subject, object) {
        return { subject, predicate: TripleUtil_1.allTypesAndSuperTypesPath, object };
    }
    resolveValueToTerm(value) {
        if ((0, Util_1.isUrl)(value)) {
            return data_model_1.default.namedNode(value);
        }
        return (0, TripleUtil_1.valueToLiteral)(value);
    }
    createOrderQueryData(subject, order) {
        if (!order) {
            return { triples: [], orders: [], variables: [] };
        }
        return Object.entries(order).reduce((obj, [property, direction]) => {
            const orderQueryData = this.createOrderQueryDataForProperty(subject, property, direction);
            obj.orders = [...obj.orders, orderQueryData.order];
            obj.variables.push(orderQueryData.variable);
            if (orderQueryData.triple) {
                obj.triples.push(orderQueryData.triple);
            }
            return obj;
        }, { triples: [], orders: [], variables: [] });
    }
    createOrderQueryDataForProperty(subject, property, direction) {
        if (property === 'id') {
            return {
                triple: undefined,
                order: {
                    expression: subject,
                    descending: direction === 'DESC' || direction === 'desc',
                },
                variable: subject,
            };
        }
        const variable = this.createVariable();
        return {
            triple: {
                subject,
                predicate: data_model_1.default.namedNode(property),
                object: variable,
            },
            order: {
                expression: variable,
                descending: direction === 'DESC' || direction === 'desc',
            },
            variable,
        };
    }
    createRelationsQueryData(subject, relations) {
        if (!relations) {
            return { variables: [], triples: [], graphTriples: [] };
        }
        return Object.entries(relations).reduce((obj, [property, value]) => {
            const variable = this.createVariable();
            obj.variables.push(variable);
            obj.triples.push({
                subject,
                predicate: data_model_1.default.namedNode(property),
                object: variable,
            });
            if (typeof value === 'object') {
                const subRelationsQueryData = this.createRelationsQueryData(variable, value);
                obj.variables = [...obj.variables, ...subRelationsQueryData.variables];
                obj.graphTriples = [
                    ...obj.graphTriples,
                    ...subRelationsQueryData.triples,
                    ...subRelationsQueryData.graphTriples,
                ];
            }
            return obj;
        }, { variables: [], triples: [], graphTriples: [] });
    }
    createVariable() {
        return data_model_1.default.variable(this.variableGenerator.getNext());
    }
}
exports.SparqlQueryBuilder = SparqlQueryBuilder;
//# sourceMappingURL=SparqlQueryBuilder.js.map