"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SparqlUpdateBuilder = void 0;
const rmlmapper_js_1 = require("@comake/rmlmapper-js");
const data_model_1 = __importDefault(require("@rdfjs/data-model"));
const TripleUtil_1 = require("../../util/TripleUtil");
const Util_1 = require("../../util/Util");
const Vocabularies_1 = require("../../util/Vocabularies");
const VariableGenerator_1 = require("./VariableGenerator");
class SparqlUpdateBuilder {
    constructor(args) {
        this.variableGenerator = new VariableGenerator_1.VariableGenerator();
        this.setTimestamps = args?.setTimestamps ?? false;
    }
    buildUpdate(entityOrEntities) {
        const entities = (0, Util_1.ensureArray)(entityOrEntities);
        const { insertions, deletions } = this.entitiesToGraphDeletionsAndInsertions(entities);
        return this.buildUpdateWithInsertionsAndDeletions(deletions, insertions);
    }
    buildDelete(entityOrEntities) {
        const entities = (0, Util_1.ensureArray)(entityOrEntities);
        const deletions = this.entitiesToGraphDeletions(entities);
        return this.buildUpdateWithInsertionsAndDeletions(deletions);
    }
    buildDeleteAll() {
        const updates = [{
                updateType: 'deletewhere',
                delete: [{
                        type: 'bgp',
                        triples: [{ subject: TripleUtil_1.subjectNode, predicate: TripleUtil_1.predicateNode, object: TripleUtil_1.objectNode }],
                    }],
            }];
        return this.sparqlUpdate(updates);
    }
    entitiesToGraphDeletionsAndInsertions(entities) {
        return entities.reduce((obj, entity) => {
            const entityGraphName = data_model_1.default.namedNode(entity['@id']);
            obj.deletions.push(this.entityToGraphDeletion(entityGraphName));
            obj.insertions.push(this.entityToGraphInsertion(entityGraphName, entity));
            return obj;
        }, { insertions: [], deletions: [] });
    }
    entitiesToGraphDeletions(entities) {
        return entities.reduce((arr, entity) => {
            const entityGraphName = data_model_1.default.namedNode(entity['@id']);
            arr.push(this.entityToGraphDeletion(entityGraphName));
            return arr;
        }, []);
    }
    entityToGraphInsertion(entityGraphName, entity) {
        const triples = this.entityToTriples(entity, entityGraphName);
        return {
            type: 'graph',
            name: entityGraphName,
            triples,
        };
    }
    entityToGraphDeletion(entityGraphName) {
        const subject = data_model_1.default.variable(this.variableGenerator.getNext());
        const predicate = data_model_1.default.variable(this.variableGenerator.getNext());
        const object = data_model_1.default.variable(this.variableGenerator.getNext());
        return {
            type: 'graph',
            name: entityGraphName,
            triples: [{ subject, predicate, object }],
        };
    }
    entityToTriples(entity, subject) {
        const entityTriples = Object.entries(entity).reduce((triples, [key, value]) => {
            const values = (0, Util_1.ensureArray)(value);
            if (key !== '@id') {
                if (key === '@type') {
                    return [
                        ...triples,
                        ...this.buildTriplesWithSubjectPredicateAndIriValue(subject, TripleUtil_1.rdfTypeNamedNode, values),
                    ];
                }
                return [
                    ...triples,
                    ...this.buildTriplesForSubjectPredicateAndValues(subject, key, values),
                ];
            }
            return triples;
        }, []);
        if (this.setTimestamps && subject.termType === 'NamedNode') {
            if (!entity[Vocabularies_1.DCTERMS.created]) {
                entityTriples.push({ subject, predicate: TripleUtil_1.created, object: TripleUtil_1.now });
            }
            entityTriples.push({ subject, predicate: TripleUtil_1.modified, object: TripleUtil_1.now });
        }
        return entityTriples;
    }
    buildTriplesForSubjectPredicateAndValues(subject, predicate, values) {
        const predicateTerm = data_model_1.default.namedNode(predicate);
        return values.flatMap((value) => this.buildTriplesWithSubjectPredicateAndValue(subject, predicateTerm, value));
    }
    buildTriplesWithSubjectPredicateAndIriValue(subject, predicate, values) {
        return values.map((valueItem) => ({
            subject,
            predicate,
            object: data_model_1.default.namedNode(valueItem),
        }));
    }
    buildTriplesWithSubjectPredicateAndValue(subject, predicate, value) {
        const isObject = typeof value === 'object';
        if (isObject) {
            if ('@value' in value) {
                return [{ subject, predicate, object: this.jsonLdValueObjectToLiteral(value) }];
            }
            const isReferenceObject = '@id' in value;
            const isBlankNodeReferenceObject = !isReferenceObject || value['@id'].startsWith('_:');
            if (isBlankNodeReferenceObject) {
                return this.buildTriplesForBlankNode(subject, predicate, value);
            }
            if (isReferenceObject) {
                return [{ subject, predicate, object: data_model_1.default.namedNode(value['@id']) }];
            }
        }
        return [{ subject, predicate, object: (0, TripleUtil_1.valueToLiteral)(value) }];
    }
    jsonLdValueObjectToLiteral(value) {
        if (typeof value['@value'] === 'object') {
            return data_model_1.default.literal(JSON.stringify(value['@value']), rmlmapper_js_1.RDF.JSON);
        }
        if (value['@language']) {
            return data_model_1.default.literal(value['@value'], value['@language']);
        }
        if (value['@type']) {
            return data_model_1.default.literal(value['@value'].toString(), value['@type']);
        }
        return (0, TripleUtil_1.valueToLiteral)(value['@value']);
    }
    buildTriplesForBlankNode(subject, predicate, value) {
        const blankNode = data_model_1.default.blankNode(this.variableGenerator.getNext());
        return [
            { subject, predicate, object: blankNode },
            ...this.entityToTriples(value, blankNode),
        ];
    }
    buildUpdateWithInsertionsAndDeletions(deletions, insertions = []) {
        const updates = this.createUpdatesFromInsertionsAndDeletions(deletions, insertions);
        return this.sparqlUpdate(updates);
    }
    createUpdatesFromInsertionsAndDeletions(deletions, insertions) {
        const updates = [];
        if (deletions.length > 0) {
            updates.push({
                updateType: 'deletewhere',
                delete: deletions,
            });
        }
        if (insertions.length > 0) {
            const insert = { updateType: 'insert', insert: insertions };
            if (this.setTimestamps) {
                insert.where = [this.selectNow()];
            }
            updates.push(insert);
        }
        return updates;
    }
    sparqlUpdate(updates) {
        return {
            type: 'update',
            prefixes: {},
            updates,
        };
    }
    selectNow() {
        return {
            type: 'query',
            queryType: 'SELECT',
            prefixes: {},
            variables: [TripleUtil_1.now],
            where: [{
                    type: 'bind',
                    variable: TripleUtil_1.now,
                    expression: {
                        type: 'operation',
                        operator: 'now',
                        args: [],
                    },
                }],
        };
    }
}
exports.SparqlUpdateBuilder = SparqlUpdateBuilder;
//# sourceMappingURL=SparqlUpdateBuilder.js.map